classdef Network
% Bayesian Network Class

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

properties
    
VarNames % Variable names
ClassNames % Variable class names
Labels % Variable level labels

% optional properties of the BN

Description % A string describing the network
UserData % Additional information to be associated with the network

end

properties(SetAccess = private)

G % The DAG as a NxN sparse connectivity matrix
P % Conditional probability distribution of each node given its parents

end

methods

function net = Network(G, P, varargin)
%NETWORK Create a discrete bayesian network
%
%   NET = NETWORK(G, P) creates a discrete bayesian network with DAG G and
%   conditional probability distribution P. P should be a cell array, of
%   length M, of ND-arrays, where M is the number of nodes in G. Each cell
%   P{i} should be an L-by-L1...-by-LK array with the probabilities of each
%   of the L levels of the i-th variable given each combination of the
%   levels of its K parents. All variables are assumed to be nominal and
%   they are named as 'Var1', ..., 'VarM' and the levels of each variable 
%   are labeled as 'Value1', ... , 'ValueL'.
%
%   NET = NETWORK(G, P, 'Param1',VAL1,'Param2',VAL2,...) specifies
%   additional parameter name/value pairs chosen from the following:
%
%      'VarNames'       A cell array of strings, of length M and without
%                       duplicates, specifying the variable names. Default
%                       is 'Var1', ..., 'VarN'.
%
%      'ClassNames'     A cell array of strings, of length M, of the class
%                       names of the variables. Valid class names are 
%                       'nominal' and 'ordinal'. A single class name can
%                        also be given to be used for all variables. 
%                       Default is 'nominal'.   
%
%      'Labels'         A cell array of length M. Each cell should be a
%                       cell array of strings, of length L, containing
%                       the level labels of the corresponding variable,
%                       where L is the size of the first dimension of the
%                       corresponding conditional probability distribution 
%                       ND-array. If the corresponding variable is ordinal,
%                       the order by which the labels are given specifies
%                       the order of the corresponding levels. Default is
%                       'Value1', ... , 'ValueL' for each variable. 
%
%      'Description'    A string describing the Network. Default is [].
%
%      'UserData'       Any variable containing additional information to
%                       be associated with the network.
%
%   Example:
%
%       See ORG.MENSXMACHINA.BNET.WEATHER function code.
%
%   See also: ORG.MENSXMACHINA.BNET.BIFREAD, ORG.MENSXMACHINA.BNET.DSLREAD.

net.G = G;
net.P = P;

% Process optional arguments and do error checking
okargs =   {'VarNames', 'ClassNames', 'Labels', 'Description' 'UserData'};
defaults = {[], [], [], [], []};
[   eid emsg ...
    net.VarNames net.ClassNames net.Labels ...
    net.Description net.UserData ...
] = org.mensxmachina.commons.getargs(okargs, defaults, varargin{:});

if ~isempty(eid)
   error(sprintf('bn:Network:%s', eid), emsg);
end

end       

% public getters / setters

function net = set.Description(net, Description)

assert( isempty(Description) || ischar(Description), 'bn:Network:BadDescription', 'Description must be a string.' );

net.Description = Description;

end


function net = set.ClassNames(net, ClassNames)
    
if isempty(ClassNames)

    % set all class names to nominal
    ClassNames = repmat({'nominal'}, 1, size(net.G, 1));

else

    good_classnames = true;

    if iscellstr(ClassNames) % a cell array of strings is given

        % check if class names are given for all variables
        if length(ClassNames) == size(net.G, 1)

            % check if all classnames are 'nominal' or 'ordinal'
            good_classnames = ...
                all(cellfun(@(ClassName) ismember(ClassName, {'nominal', 'ordinal'}), ClassNames));

        else
            good_classnames = false;
        end

    elseif ischar(ClassNames) % a single string is given

        % check the given class name is 'nominal' or 'ordinal'
        if ismember(ClassNames, {'nominal', 'ordinal'})
            % replicate the class name for all variables
            ClassNames = repmat({ClassNames}, 1, size(net.G, 1));
        else
            good_classnames = false;
        end

    else
        good_classnames = false;
    end

    assert( good_classnames, ...
        'bn:Network:BadClassNames', ...
        [ 'ClassNames must be either ''nominal'' or ''ordinal''' ...
          'or a cell array of ''ordinal'' or ''nominal'' ' ...
          'whose length is equal to the number of nodes in the Network.' ] ...
        );

end

net.ClassNames = ClassNames;

end

function net = set.Labels(net, Labels)

if isempty(Labels)

    % produce labels 'Value1', 'Value2', ... , 'ValueN'
    Labels = arrayfun(@(i) {sprintf('Value%d', i)}, ...
        1:size(net.G, 1), 'UniformOutput', false);
    
else

    good_labels = ...
        iscell(Labels) && ... % a cell array
        length(Labels) == size(net.G, 1) && ... % of valid size
        all( arrayfun(@(i) ...
                iscellstr(Labels{i}) && ... % a cell array of strings
                isequal(sort(Labels{i}), unique(Labels{i})), ... % no duplicates
                1:length(Labels)) );

    assert( good_labels, ...
        'bn:Network:BadLabels', ...
        ['Labels should be a cell array of length M, where M is the ' ...
         'number of nodes in the network. Each cell should be a cell ' ...
         'array of strings, of length L and without duplicates, ' ...
         'containing the level labels of the corresponding variable, '...
         'where L is the size of the first dimension of the ' ...
         'corresponding conditional probability distribution ND-array'] ...
        );

end

net.Labels = Labels;

end

function net = set.VarNames(net, VarNames)

if isempty(VarNames)

    % produce variable names 'Var1', 'Var2', ... , 'VarN'
    VarNames = arrayfun(@(i) {sprintf('Var%d', i)}, ...
        1:size(net.G, 1), 'UniformOutput', false);
    
else

    if iscellstr(VarNames)
        if isempty(VarNames)
            good_varnames = true;
        else
            good_varnames = ...
                length(VarNames) == size(net.G, 1) && ... % valid length
                isequal(sort(VarNames), unique(VarNames)); % no duplicates
        end
    else
        good_varnames = false;
    end

    assert( good_varnames, ...
        'bn:Network:BadVarNames', ...
        ['VarNames must either an empty cell array or ' ...
         'a cell array of unique valid variable names, ' ...
         'of length equal to the number of nodes.'] ...
        );

end

net.VarNames = VarNames;

end

% private setters / getters

function net = set.G(net, G)

assert( ...
    issparse(G) && graphisdag(G), ...
    'bn:Network:BadDAG', 'G must be, well, a G.');

net.G = G;

end

function net = set.P(net, P)
  
%{    
( ... % of valid size
    ( numel(parents(net, i)) == 0 && isequal( size(net.P{i}), [ nlevels(net, i) 1 ]) ) || ...
    isequal(size(net.P{i}), [nlevels(net, i) nlevels(net, parents(net, i))]) ...
) && ... 
%} 

good_cpd = ...
    iscell(P) && ...
    length(P) == size(net.G, 1) && ...
    all( ...
        arrayfun(@(i) ...
            isreal(P{i}) && ... % real array
            ( ...  % correct # of dimensions
                ( numel(parents(net, i)) == 0 && ndims(P{i}) == 2) || ...
                ( numel(parents(net, i)) > 0 && ndims(P{i}) == length(parents(net, i)) + 1) ...
            ) && ... 
            all(abs(reshape(sum(P{i}, 1), 1, [])-1) < 1e-4), ... % probabilities sum to 1
            1:length(P)));
        
assert( good_cpd, ...
    'bn:Network:BadP', ...
    ['P must be a cell array of length equal to the number of nodes ' ...
     'in the network. Each cell must be a real ND-array with number of ' ...
     'dimensions equal to the number of parents of that node plus one, ' ...
     'or equal to 2 if the node is a root node. The sums along the ' ...
     '1st dimension must all be 1.'] ...
    );

net.P = P;

end

% other methods

function skeleton = skeleton(net)
%SKELETON Get the skeleton of a bayesian network
skeleton = org.mensxmachina.commons.graph.skeleton(net.G);
end

function parents = parents(net, i)
%PARENTS Get the list of parents of a bayesian network node      
parents = org.mensxmachina.commons.graph.parents(net.G, i);
end

function nlevels = nlevels(net, j)
%NLEVELS Get the number of levels of a bayesian network node
nlevels = arrayfun(@(j) length(net.Labels{j}), j);
end

function view(net)
%VIEW Draw figure of the network graph

view(biograph(net.G, net.VarNames)); 
    
end

function r = rnd(net, n)
% rnd Random dataset from the distribution of a bayesian network
%
% R = RND(NET, N) returns a random dataset R of N observations
% from the distribution of NET Network.

fprintf('\nSampling...\n');

% get a topological order of the nodes
order = graphtopoorder(net.G); 

levels = zeros(n, size(net.G, 1));

k = 0;

for j = order % for each node in this order
    
    k = k + 1;
    
    fprintf('\nCreating column #%d, ''%s''...\n', k, net.VarNames{j});
    
    % get values of the node
    % according to its P given its parents

    % get the parents of the node
    parents_j = parents(net, j);

    % probability (categorical) distributions of the node
    levels_parents = num2cell( levels(:, parents_j) );
    P = cell2mat(arrayfun(@(i) net.P{j}( :, levels_parents{i, :} )', (1:n)', 'UniformOutput', false));
    
    % get level counts from this distribution
    levelcounts = mnrnd(1, P);
    
    % find corresponding levels
    [row col] = find(levelcounts);  
    levels(row', j) = col';
    
end           

% initialize a cell array of dataset columns
dataset_columns = cell(1, size(net.G, 1));

for j = order % for each node

    if strcmp(net.ClassNames{j}, 'ordinal')
        % construct an ordinal column

        dataset_columns{j} = ...
            ordinal( levels(:, j), net.Labels{j}, 1:nlevels(net, j) );

    else % construct a nominal column

        dataset_columns{j} = ...
            nominal( levels(:, j), net.Labels{j}, 1:nlevels(net, j) );

    end

end

if isempty(net.VarNames)
    params = {};
else
    params = [ 'VarNames' {net.VarNames} ];
end

% create a dataset consisting of the created columns
r = dataset(dataset_columns{:}, params{:});

end

end

methods(Static)

% save / load

function obj = loadobj(s)

obj = org.mensxmachina.bnet.Network(s.G, s.P, 'VarNames', s.VarNames, ...
    'Labels', s.Labels, 'ClassNames', s.ClassNames, ...
    'Description', s.Description, 'UserData', s.UserData);

end

function s = saveobj(obj)

s.G = obj.G;
s.P = obj.P;
s.VarNames = obj.VarNames;
s.Labels = obj.Labels;
s.ClassNames = obj.ClassNames;
s.Description = obj.Description;
s.UserData = obj.UserData;

end

end

end