function net = bifread(filename, ClassNames)
%BIFREAD Read Bayesian Network from BIF file
%
%   NET = BIFREAD(FILENAME) reads a BIF format file FILENAME, returning
%	the data in the file as a network object with all nodes nominal.
%   BIFREAD calls the web service at
%   http://www.digitas.harvard.edu/cgi-bin/ken/bif2bnt in order to read the
%   file, so the computer must be connected to the Internet.
%
%   NET = BIFREAD(FILENAME, CLASSNAMES) reads a BIF format file FILENAME,
%   returning the data in the file as a network object whose node classes
%   are defined in CLASSNAMES. CLASSNAMES can be 'nominal' or 'ordinal',
%   resulting in all nodes to be of that class, or a cell array of
%   'nominal' or 'ordinal' defining the class of each node separately.
%
%   Example:
%
%       alarm_net = org.mensxmachina.bifread('alarm.bif');
%
%   See also: ORG.MENSXMACHINA.BNET.NETWORK, ORG.MENSXMACHINA.BNET.DSLREAD

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

dot_pos = strfind(filename, '.');
last_dot_pos = dot_pos(end);

% determine the filename of the cached BNT code
cache_filename = [filename(1:last_dot_pos-1) '_bifread_cache.txt' ];

if exist(cache_filename, 'file')
    
    % read BNT code from cache
    bnt_code = fileread(cache_filename);
    
else % call the web service
    
    % open BIF file
    bif = fileread(filename);

    % get BNT code
    bnt_code = urlread( ...
        'http://www.digitas.harvard.edu/cgi-bin/ken/bif2bnt', 'post', ...
        {'bif_textarea', bif} );

    % cache it
    fid = fopen(cache_filename, 'w');
    fwrite(fid, bnt_code);
    fclose(fid);
    
end

% evaluate the code
eval(bnt_code);

% set the node names of the network
bnet.names = fieldnames(node);

% get the labels of each node
value = cellfun(@(value_i) {value_i'}, value);

% create arguments for network initilization

DAG = sparse(bnet.dag);

% get probability TABLES from bnet's CPD
CPD = cellfun(@(CPD) {get_field(CPD, 'cpt')}, bnet.CPD);

% bring the parent dimensions of each CPT to the end
CPD = cellfun(@(CPD) {shiftdim(CPD, ndims(CPD)-1)}, CPD);

VarNames = bnet.names';
Labels = value;

% create the network object 
net = org.mensxmachina.bnet.Network(DAG, CPD, 'VarNames', VarNames, 'Labels', Labels, 'ClassNames', ClassNames);

end

% dummy functions to emulate BNT objects creation
% without adding BNT to path

function bnet = mk_bnet(dag, node_sizes)
% TABULAR_CPD emulates the BNT mk_bnet() function
% Instead of a BNT network struct, it returns a struct
% with the 'dag', 'node_sizes' and 'CPD' fields only.

bnet.dag = dag;
bnet.node_sizes = node_sizes(:)';
bnet.CPD = cell(1, numel(node_sizes));

end

function CPD = tabular_CPD(bnet, self, cpt)
% TABULAR_CPD emulates the constructor of a BNT tabular_CPD class
% Instead of a tabular_CPD object, it returns a struct
% with the 'cpt' field only.

% get node sizes from the BNT network struct
ns = bnet.node_sizes;

% get parents of this node
ps = org.mensxmachina.commons.graph.parents(bnet.dag, self);

% get the sizes of the parents of the node
% and its size also; this will be the size of the CPT ND-array
fam_sz = ns([ps self]);

if numel(fam_sz) == 1 % if it's a marginal node
    fam_sz = [1 fam_sz];
end

% reshape the given cpt vector into into a ND-array
% where the first dimensions correspond to the parents
% and the last to the current node
CPD.cpt = reshape(cpt, fam_sz);

end

function value = get_field(CPD, fieldname)
    value = CPD.(fieldname);
end