function net = dslread( graphfilename, nodesfilename, dcfilename )
%DSLREAD Read DSL network files
%
%   NET = DSLREAD(GRAPHFILE, NODESFILE, DCFILE) reads the DSL (Discovery 
%   Systems Laboratory, Vanderbilt University) network graph file 
%   GRAPHFILE, nodes file NODESFILE and domain counts file DCFILE into a
%   Network object.
%
%   Example:
% 
%       pigs_net = org.mensxmachina.bnet.dslread( ...
%           'dsl_ce_pigs_graph', 'dsl_ce_pigs_nodes', 'dsl_ce_pigs_dc');
%
%   See also: ORG.MENSXMACHINA.BNET.NETWORK, ORG.MENSXMACHINA.BNET.BIFREAD

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

load(graphfilename, 'graph');
load(nodesfilename, 'nodes');
load(dcfilename, 'domain_counts');

% DSL network format:
% nodes{i}
%         .parents = Parents of the node i
%         .cpt     = Cumulative probability table
%         .name    = Name of the node

numNodes = numel(nodes);

DAG = sparse(graph);

CPD  = cell(1, numNodes);

for j=1:numNodes
    
    if isempty(nodes{j}.parents)
        CPD{j} = nodes{j}.cpt;
    else    
        
        [B order] = sort(nodes{j}.parents);
        CPD{j} = permute(nodes{j}.cpt, [ 1 order+1 ]);
        
    end
    
end

VarNames = cellfun(@(node) {node.name}, nodes);
Labels = arrayfun(@(j) arrayfun(@(k) sprintf('%s_%d', VarNames{j}, k), 1:domain_counts(j), 'UniformOutput', false), 1:numNodes, 'UniformOutput', false);

% create the network object
net = org.mensxmachina.bnet.Network(DAG, CPD, 'VarNames', VarNames, 'Labels', Labels, 'ClassNames', 'nominal');

end
