function [ind pMax pMaxStat] = mmpc(a, t, varargin)
%MMPC Learn parents and children using the MMPC algorithm
% 
%   MMPC is called just like MMPCBAR but accepts the MMSKELETON parameter
%   name/value pairs instead, with the differences that MMSKELETON
%   'includeNpc' parameter corresponds to MMPC 'includeNpc' parameter and
%   there is no MMPC 'sc' parameter (symmetry correction is performed,
%   by definition). MMPC returns the parents and children set instead
%   of the extended and children set.
%
%   Example:
%
%       % load ALARM network
%       load mxm_bnet_alarm_net; % loads 'net' Network
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % learn parents and children set of variable #37
%       pc = org.mensxmachina.bnet.mmpc(a, 37);
%
%   See also
%       ORG.MENSXMACHINA.BNET.MMPCBAR, ORG.MENSXMACHINA.BNET.mmskeleton.

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

% start copy-paste from mmskeleton
% change p.FunctionName
% add p.addRequired('t', ... line
% remove p.addParamValue('sc', ... line
% change p.parse(a, varargin{:}); to p.parse(a, t, varargin{:});
% rename includeNpc with includeNpc 

import org.mensxmachina.bnet.*;
import org.mensxmachina.commons.stats.*;

p = inputParser;  % Create an instance of the inputParser class
p.FunctionName = 'ORG.MENSXMACHINA.BNET.MMPC';

p.addRequired('a', @iscategoricaldataset);
p.addRequired('t', @(t) ischar(t) || (isscalar(t) && org.mensxmachina.commons.ispositiveinteger(t)));

p.addParamValue('alpha', 0.05, @(alpha) isscalar(alpha) && alpha >= 0 && alpha <= 1);
p.addParamValue('omin', 5, @(omin) isscalar(omin) && omin>=0 && isreal(omin) && omin==round(omin));
p.addParamValue('nsmax', 10, @(nsmax) isscalar(nsmax) && nsmax>=0 && isreal(nsmax) && nsmax==round(nsmax));
p.addParamValue('testFun', @chi2testopt, @(testFun) isa(testFun, 'function_handle'));
p.addParamValue('rl', 1, @(rl) isreal(rl) && isscalar(rl) && ismember(rl, [0 1 2 3]));
p.addParamValue('includeNpc', false, @islogical);
p.addParamValue('verbose', false, @islogical);
p.addParamValue('subs', [], @(subs) isnumeric(subs) && ndims(subs) == 2 );
p.addParamValue('nLevels', [], @(nLevels) isnlevels(a, nLevels) );

p.parse(a, t, varargin{:});

Params = p.Results;

subs = p.Results.subs;
nLevels = p.Results.nLevels;

Params = rmfield(Params, {'subs', 'nLevels'});

if isempty(subs)
    % convert dataset to subscripts
    subs = double(a);
end 

if isempty(nLevels)
    % get # of levels
    nLevels = nlevels(a);
end

% get V, the set of all variables
v = 1:size(a, 2);

nv = numel(v);

% initialize EPC and NEPC, the sets of EPC(T) and NEPC(T) for each T in V
if Params.rl >= 1
    
    Params.Nepc.ind = cell(1, nv);
    Params.Nepc.pMax = cell(1, nv);
    Params.Nepc.pMaxStat = cell(1, nv);
    
    if Params.rl >= 2
    
        Params.Epc.ind = cell(1, nv);
        Params.Epc.pMax = cell(1, nv);
        Params.Epc.pMaxStat = cell(1, nv);

        % initialize MEPC, the set of MEPC(T) for each T in V
        if Params.rl == 3
            Params.Mepc.ind = cell(1, nv);    
        else
            Params.Mepc = []; 
        end
    
    else
        Params.Epc = [];
        Params.Mepc = []; 
    end
    
else
    Params.Nepc = [];
    Params.Epc = [];    
    Params.Mepc = [];
end

% end copy-paste from mmskeleton

% learn EPC(T)
fprintf('\nLearning EPC(%d)...\n', t);

Params.includeNepc = Params.includeNpc;

% call optimized version of MMPC-bar
[epc_t_ind pMax_t pMaxStat_t Params.Nepc Params.Epc Params.Mepc] = ...
    org.mensxmachina.bnet.mmpcbaropt(subs, nLevels, t, Params);

if ~Params.includeNpc
    
    % sort EPC(T)
    [epc_t_ind epc_t_ind_sort_ind] = sort(epc_t_ind);
    pMax_t = pMax_t(epc_t_ind_sort_ind);
    pMaxStat_t = pMaxStat_t(epc_t_ind_sort_ind);

    % initialize PC(T)
    ind = [];
    pMax = [];
    pMaxStat = [];

    for x_ind_in_epc_t = 1:numel(epc_t_ind) % for each variable X in EPC(T)

        % let it be X
        x = epc_t_ind(x_ind_in_epc_t);

        % learn EPC(X)
        fprintf('\nLearning EPC(%d)...\n', x);

        % call optimized version of MMPC-bar
        [epc_x_ind pMax_x pMaxStat_x Params.Nepc Params.Epc Params.Mepc] = ...
            org.mensxmachina.bnet.mmpcbaropt(subs, nLevels, x, Params);

        % find index of T in EPC(X)
        t_ind_in_epc_x = find(epc_x_ind == t, 1);

        if ~isempty(t_ind_in_epc_x) % if T is also in EPC(X)

            % add X to PC(T)
            ind = [ind x];      

            if pMax_t(x_ind_in_epc_t) >= pMax_x(t_ind_in_epc_x)

                pMax = [pMax pMax_t(x_ind_in_epc_t)];
                pMaxStat = [pMaxStat pMaxStat_t(x_ind_in_epc_t)];

            else

                pMax = [pMax pMax_x(t_ind_in_epc_x)];
                pMaxStat = [pMaxStat pMaxStat_x(t_ind_in_epc_x)];

            end

        end

    end
    
else

    % initialize PC(T)
    ind = sparse(1, nv);
    pMax = pMax_t;
    pMaxStat = pMaxStat_t;

    for x = find(epc_t_ind) % for each variable X in EPC(T)

        % learn EPC(X)
        fprintf('\nLearning EPC(%d)...\n', x);

        % call optimized version of MMPC-bar
        [epc_x_ind pMax_x pMaxStat_x Params.Nepc Params.Epc Params.Mepc] = ...
            org.mensxmachina.bnet.mmpcbaropt(subs, nLevels, x, Params);

        if epc_x_ind(t) % if T is also in EPC(X)

            % add X to PC(T)
            ind(x) = true;   

            if pMax_t(x) >= pMax_x(t)

                pMax(x) = pMax_t(x);
                pMaxStat(x) = pMaxStat_t(x);

            else

                pMax(x) = pMax_x(t);
                pMaxStat(x) = pMaxStat_x(t);

            end

        end

    end
    
end

end