function [ind pMax pMaxStat] = mmpcbar(a, t, varargin)
%MMPCBAR Learn extended parents and children using MMPC-bar algorithm
% 
%   IND = MMPCBAR(A, T) learns the variables in dataset A that form the
%   extended parents and children (EPC) set of the variable specified by T
%   in A and returns the linear indices of those variables in row vector
%   IND. A must be a  dataset with categorical only variables and T can be 
%   a positive integer or a variable name in A.
%
%   [IND, PMAX] = MMPCBAR(A, T) also returns a row vector PMAX of the same
%   length as IND containing the maximum p-values among the hypothesis
%   tests of conditional independence of each variable in EPC(T).
%
%   [IND, PMAX, PMAXSTAT] = MMPCBAR(A, T) also returns a row vector 
%   PMAXSTAT of the same length as PMAX and IND containing the G-square
%   statistics corresponding to the maximum p-values in PMAX.
%
%   [...] = MMPCBAR(A, T, 'Param1',VAL1,'Param2',VAL2,...) specifies
%   additional parameter name/value pairs chosen from the following:
%
%      'alpha'          A value between 0 and 1 that specifies the
%                       significance level as (100*ALPHA)% for the 
%                       hypothesis tests of conditional independence that 
%                       are performed. Default is 0.05.
%      'omin'           A non-negative integer that specifies the minimum
%                       average observed count per cell (bin) required for
%                       a test to be performed. MMPCBAR does not perform a
%                       test if the average observed count per cell is 
%                       smaller than OMIN and accepts the null hypothesis
%                       if the conditioning set is the empty set otherwise
%                       it rejects it. Default is 5.
%      'nsmax'          A non-negative integer that specifies the maximum
%                       cardinality of a conditioning set. MMPCBAR does not
%                       perform hypothesis tests of conditional
%                       independence with T conditioning on sets with 
%                       cardinality greater than NSMAX. Default is 10.
%      'testFun'        A handle to a function that performs a hypothesis
%                       test of conditional independence. Default is
%                       CHI2TESTOPT, which
%                       performs a G-square test of conditional
%                       independence. See that function's documentation
%                       for the requirements of such a testing function.
%      'includeNepc'    A logical scalar that specifies whether information
%                       about non- extended parents and children is
%                       returned. If logical 1 (true), IND is a 1-by-M
%                       sparse vector of logical indices, where M is the
%                       number of variables in A, and PMAX and PMAXSTAT are
%                       1-by-M vectors that contain the maximum p-values 
%                       and their corresponding statistics for all 
%                       variables in A. A maximum p-value corresponding to
%                       a variable in NEPC is the p-value of the last
%                       testing of conditional independence with T, which 
%                       determined its status as a non- extended parent or
%                       child, and it is either > ALPHA or NaN, being the
%                       latter in the case that the test of conditional
%                       independence with T given the empty set wasn't
%                       performed and subsequently independence was
%                       accepted. Default is false.
%      'verbose'        A logical scalar that specifies whether messages
%                       are displayed during the execution of the function.
%                       Default is false.
%   Example:
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % learn extended parents and children set of variable #37
%       epc_alarm_37 = org.mensxmachina.bnet.mmpcbar(a, 37)
%
%   See also
%       ORG.MENSXMACHINA.BNET.MMPC, ORG.MENSXMACHINA.BNET.MMSKELETON.

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

import org.mensxmachina.commons.stats.*;

p = inputParser;  % Create an instance of the inputParser class
p.FunctionName = 'ORG.MENSXMACHINA.BNET.MMPCBAR';

p.addRequired('a', @iscategoricaldataset);
p.addRequired('t', @(t) ischar(t) || (isscalar(t) && org.mensxmachina.commons.ispositiveinteger(t)));

p.addParamValue('alpha', 0.05, @(alpha) isscalar(alpha) && alpha >= 0 && alpha <= 1);
p.addParamValue('omin', 5, @(omin) isscalar(omin) && omin>=0 && isreal(omin) && omin==round(omin));
p.addParamValue('nsmax', 10, @(nsmax) isscalar(nsmax) && nsmax>=0 && isreal(nsmax) && nsmax==round(nsmax));
p.addParamValue('testFun', @chi2testopt, @(testFun) isa(testFun, 'function_handle'));
p.addParamValue('includeNepc', false, @islogical);
p.addParamValue('verbose', false, @islogical);
p.addParamValue('subs', [], @(subs) isnumeric(subs) && ndims(subs) == 2 );
p.addParamValue('nLevels', [], @(nLevels) isnlevels(a, nLevels) );

p.parse(a, t, varargin{:});

if ischar(t)
    % find index of the variable given its name
    varNames = get(a, 'VarNames');
    t = find(strcmp(varNames, t), 1);
end

Params = p.Results;

subs = p.Results.subs;
nLevels = p.Results.nLevels;

Params = rmfield(Params, {'subs', 'nLevels'});

if isempty(subs)
    % convert dataset to subscripts
    subs = double(a);
end 

if isempty(nLevels)
    % get # of levels
    nLevels = nlevels(a);
end

Params.Nepc = [];
Params.Epc = [];
Params.Mepc = [];

% call optimized version of MMPC-bar
[ind pMax pMaxStat] = ...
    org.mensxmachina.bnet.mmpcbaropt(subs, nLevels, t, Params);
    
end  