function [ind pMax pMaxStat Nepc Epc Mepc] = mmpcbaropt(subs, nLevels, t, Params)
%MMPCBAROPT Learn extended parents and children using MMPC-bar algorithm
%
%   MMPCBAROPT is the optimized version of MMPCBAR called by MMSKELETON.
%
%   IND = MMPCBAROPT(SUBS, NLEVELS, T, PARAMS) learns the columns in matrix
%   SUBS that form the extended parents and children (EPC) set of column T
%   and returns the linear indices of those variables in row vector
%   IND. SUBS should be a matrix of levels and NLEVELS should be the number
%   of levels vector of SUBS. SUBS should not contain unknown (NaN) values.
%   PARAMS should be a struct containing the alpha, omin, nsmax, testFun,
%   includeNepc and verbose parameters, given as parameter name/value pairs
%   in function ORG.MENSXMACHINA.BNET.MMPCBAR, as fields. PARAMS should
%   also contain the following fields:
%
%      'Nepc'           A struct containing up-to-date information about
%                       the non- extended parents or children of the
%                       columns in SUBS, in order to be reused by
%                       MMPCBAROPT. The NEPC.ind field should be a 1-by-M
%                       cell array of the row vectors with the linear
%                       indices of the extended parents and children of
%                       the columns in SUBS. EPC.pMax and EPC.pMaxStat
%                       fields should be 1-by-M cell arrays of the maximum
%                       p-values and the corresponding statistics of the 
%                       extended parents and children of the columns in
%                       SUBS, and should be included as fields only if the 
%                       'includeNepc' parameter (see
%                       ORG.MENSXMACHINA.BNET.MMPCBAR documentation)
%                       is set to true. NEPC can also be the empty array
%                       ([]), implying that no reuse is desired. If [], 
%                       the 'Epc' parameter below should also be [].
%
%      'Epc'            A struct containing up-to-date information about
%                       the extended parents and children of the columns in
%                       SUBS, in order to be reused by MMPCBAROPT. The 
%                       format of the struct is similar to that of the NEPC
%                       struct above, with the exception that the pMax and
%                       pMaxStat fields should be always included. EPC can
%                       also be the empty array ([]), implying that no
%                       reuse is desired. If [], the MEPC parameter below
%                       should also be []. 
%
%      'Mepc'           A struct containing up-to-date information about
%                       the "maximal" extended parents and children of the 
%                       columns in SUBS, in order to be reused by 
%                       MMPCBAROPT. The maximal set of extended parents and
%                       children ("MEPC") is the set of extended parents
%                       and children returned by the first (forward) phase
%                       of the MMPC-bar algorithm. MEPC should only contain
%                       a single field ind, which should be a 1-by-M cell
%                       array of the row vectors with the linear indices of
%                       the maximal extended parents and children of the
%                       variables in SUBS. MEPC can also be the empty array
%                       ([]), implying that no reuse is desired.
%
%   [IND, PMAX] = MMPCBAROPT(SUBS, NLEVELS, T, PARAMS) also returns a row
%   vector PMAX of the same length as IND containing the maximum p-values 
%   among the hypothesis tests of conditional independence of each variable
%   in EPC(T).
%
%   [IND, PMAX, PMAXSTAT] = MMPCBAROPT(SUBS, NLEVELS, T, PARAMS) also 
%   returns a row vector PMAXSTAT of the same length as PMAX and IND
%   containing the G-square statistics corresponding to the maximum 
%   p-values in PMAX.  
%
%   [IND, PMAX, PMAXSTAT, EPC] = MMPCBAROPT(SUBS, NLEVELS, T, PARAMS) also
%   returns a struct EPC containing up-to-date information about the 
%   extended parent and children of the columns in SUBS. EPC is 
%   returned by a call to MMPCBAROPT from 
%   ORG.MENSXMACHINA.BNET.MMSKELETON and passed to the next call via the
%   PARAMS 'Epc' field (see above) in order to enable its reuse (see the 
%   'rl' parameter in ORG.MENSXMACHINA.BNET.MMSKELETON). If the PARAMS
%   'Epc' field is empty, EPC is also empty, i.e. no information is
%   returned.
%
%   [IND, PMAX, PMAXSTAT, EPC, NEPC] = MMPCBAROPT(SUBS, NLEVELS, T, PARAMS)
%   also returns a struct NEPC containing up-to-date information about the
%   non- extended parents and children of the columns in SUBS. NEPC is
%   returned by a call to MMPCBAROPT from ORG.MENSXMACHINA.BNET.MMSKELETON
%   and passed to the next call via the PARAMS 'Nepc' field (see above) in
%   order to enable its reuse (see the 'rl' parameter in 
%   ORG.MENSXMACHINA.BNET.MMSKELETON). If the PARAMS 'Nepc' field is empty,
%   NEPC is also empty, i.e. no information is returned.
%
%   [IND, PMAX, PMAXSTAT, EPC, NEPC, MEPC] = ...
%       MMPCBAROPT(SUBS, NLEVELS, T, PARAMS) also returns a struct MEPC
%   containing up-to-date information about the "maximal" extended parents
%   and children set of the variables in SUBS. MEPC is returned by a call
%   to MMPCBAROPT from ORG.MENSXMACHINA.BNET.MMSKELETON and passed to the
%   next call via the PARAMS 'Mepc' field (see above) in order to enable
%   its reuse (see 'rl' parameter in ORG.MENSXMACHINA.BNET.MMSKELETON). 
%   If the PARAMS 'Mepc' field is empty, MEPC is also empty, i.e. no
%   information is returned.
%
%   Example:
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % convert dataset to subscripts
%       subs = double(a);
%
%       % get number of levels
%       nLevels = org.mensxmachina.commons.stats.nlevels(a);
%
%       % create parameters struct
%       Params.alpha = 0.05;
%       Params.omin = 5;
%       Params.nsmax = 10;
%       Params.testFun = @org.mensxmachina.commons.stats.chi2testopt;
%       Params.includeNepc = false;
%       Params.verbose = false;
%       Params.Epc = [];
%       Params.Nepc = [];
%       Params.Mepc = [];
%
%       % learn extended parents and children set of variable #37
%       epc_alarm_37 = ...
%           org.mensxmachina.bnet.mmpcbaropt(subs, nLevels, 37, Params)
%
%   See also
%       ORG.MENSXMACHINA.BNET.MMPCBAR, ORG.MENSXMACHINA.BNET.MMSKELETON.

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

% get the total number of variables (|V|)
nv = size(subs, 2);

% get V, the indices of all variables
v = 1:nv;

% initialize NEPC(T) (non-EPC(T)),
% the set of variables Y such that X is -not- in EPC(Y)

if ~isempty(Params.Nepc) % RL >= 1
    
    % optimization 2.4 @ MMHC paper
    % find variables Y with computed EPC(Y) such that X is not in EPC(Y) 
    % in order not to be considered when learning EPC(X):
    % X not in EPC(Y) => X not in PC(Y) <=> Y not in PC(X)
    
    Nepc_t.ind = [];
    
    if Params.includeNepc
        Nepc_t.pMax = [];
        Nepc_t.pMaxStat = [];
    end
    
    for x=1:t-1 % for each previous variable X
        
        % find index of T in NEPC(X)
        tIndInNepc_x = find(Params.Nepc.ind{x} == t, 1);
        
        if ~isempty(tIndInNepc_x) % found
            
            % insert X in NEPC(T)
            Nepc_t.ind = [ Nepc_t.ind x ];
            
            if Params.includeNepc
            
                % get the maximum p-value
                % among the tests I(X;T|S) performed while learning EPC(X)
                % note: this is the p-value of the last I(X;T|S) test made, which
                % lead to the inclusion of T in NEPC(X)
                Nepc_t.pMax = [ Nepc_t.pMax Params.Nepc.pMax{x}(tIndInNepc_x) ];
                
                % get the corresponding statistic
                Nepc_t.pMaxStat = [ Nepc_t.pMaxStat Params.Nepc.pMaxStat{x}(tIndInNepc_x) ];
            
            end
            
        end
        
    end    

    if Params.verbose
        fprintf('\nNEPC(%d) = [ %s ]\n', t, num2str(Nepc_t.ind));
    end
    
else
    
    Nepc_t.ind = [];
    
    if Params.includeNepc
        Nepc_t.pMax = [];
        Nepc_t.pMaxStat = [];
    end
    
end

if ~isempty(Params.Epc) % RL >= 2
    
    % find variables X in R if T is in EPC(X)
    % we call these variables "possible parents and children" of T
    % and we denote them as PPC(T)
    
    Ppc_t.ind = [];
    Ppc_t.tIndInEpc = [];
    Ppc_t.assocMin = [];
    Ppc_t.assocMinStat = [];
    
    if ~isempty(Params.Mepc)
        Ppc_t.tIndInMepc = [];
    end
    
    for x=1:t-1 % for each previous variable X
        
        % find index of T in EPC(X)
        tIndInEpc_x = find(Params.Epc.ind{x} == t, 1);
        
        if ~isempty(tIndInEpc_x) % found
            
            % insert X in PPC(T)
            Ppc_t.ind = [ Ppc_t.ind x ];
            Ppc_t.tIndInEpc = [ Ppc_t.tIndInEpc tIndInEpc_x ];
            
            % set the minimum association with T
            % as the negative maximum p-value among the tests I(X;T|S) 
            % performed while learning EPC(X)
            Ppc_t.assocMin = [ Ppc_t.assocMin -Params.Epc.pMax{x}(tIndInEpc_x) ];

            % get the corresponding statistic
            Ppc_t.assocMinStat = [ Ppc_t.assocMinStat Params.Epc.pMaxStat{x}(tIndInEpc_x) ];
    
            if ~isempty(Params.Mepc) % RL == 3

                % find index of T in MEPC(X)
                Ppc_t.tIndInMepc = [ Ppc_t.tIndInMepc find(Params.Mepc.ind{x} == t, 1) ];
                
            end
            
        end
        
    end
    
    if isempty(Params.Mepc) % RL == 2
        % set for each X in PPC(T) the index of T in MEPC(X) to unknown (NaN)
        Ppc_t.tIndInMepc = NaN(1, numel(Ppc_t.ind)); 
    end
    
    if Params.verbose
        fprintf('\nPN(%d) = [ %s ]\n', t, num2str(Ppc_t.ind));
    end

else
    
    Ppc_t.ind = [];
    
end

% initialize R, the set remaining variables, to V\({T} U NEPC(T) U PPC(T))
r.ind = setdiff(v, [t Nepc_t.ind Ppc_t.ind]);

if Params.verbose
    fprintf('\nR(%d) = [ %s ]\n', t, num2str(r.ind));
end 

% initialize the minimum associations of the variables in R with T
% given all the subsets of EPC(T), to the maximum association, i.e zero
r.assocMin = zeros(1, numel(r.ind));

% initialize the corresponding statistics to the maximum statistic, i.e Inf
r.assocMinStat = Inf(1, numel(r.ind));

% none of them is in PPC(T),
% so there is no index of T in their EPC and MEPC
r.tIndInEpc = NaN(1, numel(r.ind));
r.tIndInMepc = NaN(1, numel(r.ind));

% initialize EPC(T), the set of extended parents and children of T 
Epc_t.ind = []; % initialize EPC(T) to the empty set
Epc_t.pMax = []; % maximum p-values
Epc_t.pMaxStat = []; % their statistics
Epc_t.tIndInEpc = []; % index of T in EPC(X), for each X in EPC(T)
Epc_t.tIndInMepc = []; % index of T in MEPC(X), for each X in EPC(T)

% phase I (forward): add variables to EPC(T)
[Epc_t Nepc_t] = forward(subs, nLevels,  t, Epc_t, Nepc_t, Ppc_t, r, Params);

if nargout == 6 % MEPC data should be returned

    if ~isempty(Params.Mepc) % RL == 3        
        % populate MEPC data with the current (which is the maximal) EPC(T)
        Params.Mepc.ind{t} = Epc_t.ind;        
    else        
        % no MEPC reuse, so no MEPC data are returned
        Params.Mepc = [];        
    end

    Mepc = Params.Mepc;
    
    if Params.verbose
        fprintf('\nMEPC(%d) = [ %s ]\n', t, num2str(Epc_t.ind));
    end    
    
end

% phase II (backward): remove variables from EPC(T)
[Epc_t Nepc_t] = backward(subs, nLevels, t, Epc_t, Nepc_t, Params);

% debug
%assert(isequal(sort([Epc_t.ind Nepc_t.ind t]), v));

if Params.verbose
    fprintf('\nFinal EPC(%d) = [ %s ]\n', t, num2str(Epc_t.ind));
end

if Params.includeNepc

    % convert EPC(T) linear to logical indices as the 1st output
    ind = sparse(1, nv);
    ind(Epc_t.ind) = true;
    
else
    
    % return EPC(T) indices as the 1st output
    ind = Epc_t.ind;  
    
end

if nargout >= 2 % maximum p-values must be returned
    
    if Params.includeNepc
    
        % copy all maximum p-values to the 2nd output
        % note: pMax(t) is set to 1
        pMax = ones(1, nv);
        pMax(Epc_t.ind) = Epc_t.pMax;
        pMax(Nepc_t.ind) = Nepc_t.pMax;
        
    else
        
        % copy EPC(T) maximum p-values to the 2nd output
        pMax = Epc_t.pMax;
        
    end
    
end

if nargout >= 3 % statistics must me returned
    
    if Params.includeNepc
        
        % copy stats corresponding to all maximum p-values
        % to the 3rd output
        % note: pMaxStat(t) is set to 0
        pMaxStat = zeros(1, nv);
        pMaxStat(Epc_t.ind) = Epc_t.pMaxStat;
        pMaxStat(Nepc_t.ind) = Nepc_t.pMaxStat;
        
    else
        
        % copy stats corresponding to EPC(T) maximum p-values
        % to the 3rd output
        pMaxStat = Epc_t.pMaxStat;        
        
    end
    
end

if nargout >= 4 % EPC and NEPC data must be returned

    if ~isempty(Params.Nepc) % RL >=1
        
        % add NEPC(T) data to the NEPC data
        Params.Nepc.ind{t} = Nepc_t.ind;

        if Params.includeNepc
            Params.Nepc.pMax{t} = Nepc_t.pMax;
            Params.Nepc.pMaxStat{t} = Nepc_t.pMaxStat;
        end            
        
        if ~isempty(Params.Epc) % RL >=2

            % add EPC(T) data to the EPC data
            Params.Epc.ind{t} = Epc_t.ind;
            Params.Epc.pMax{t} = Epc_t.pMax;
            Params.Epc.pMaxStat{t} = Epc_t.pMaxStat;
            
        else
            % no EPC reuse, so no EPC data are returned
            Params.Epc = [];
        end
        
    else
        % no reuse, so no NEPC and EPC data are returned
        Params.Nepc = [];
        Params.Epc = [];
    end
    
    Nepc = Params.Nepc;
    Epc = Params.Epc;
    Mepc = Params.Mepc;

end

end

function [Epc_t Nepc_t] = forward(subs, nLevels, t, Epc_t, Nepc_t, Ppc_t, r, Params)
%PHASE1 Phase I (forward): add variables to EPC(T)

while true
    
    if ~isempty(Ppc_t.ind) % there are variables left in PPC(T)
    
        % compute association of variables in PPC(T) with T given EPC(T)
        % while eliminating any variables that achieve no association
        [Ppc_t Nepc_t] = minassocupdate(subs, nLevels, t, Epc_t, Nepc_t, 'PPC', Ppc_t, Params);
    
    end
    
    if ~isempty(r.ind) % there are variables left in R
    
        % compute association of variables in R with T given EPC(T)
        % while eliminating any variables that achieve no association
        [r Nepc_t] = minassocupdate(subs, nLevels, t, Epc_t, Nepc_t, 'R', r, Params);
    
    end
    
    f = [];
        
    %{
    if ~isempty(r.ind) % there are variables left in R

        % get variable F from R
        % with the maximum minimum association with T given EPC(T)        
        [f assocmin_f statMin_f tInEpc_f tInMepc_f r] = mmheur(r);

    elseif ~isempty(Ppc_t.ind) % there are variables left in PPC(T)
    
        % get next variable F from PPC(T)
        [f assocmin_f statMin_f tInEpc_f tInMepc_f Ppc_t] = ppcnext(Ppc_t);
        
    end
   %}
    
    if ~isempty(Ppc_t.ind) % there are variables left in PPC(T)
    
        % get next variable F from PPC(T)
        %[f assocmin_f statMin_f tInEpc_f tInMepc_f Ppc_t] = ppcnext(Ppc_t);
        [f assocmin_f statMin_f tInEpc_f tInMepc_f Ppc_t] = ppcnext(Ppc_t);
        
    elseif ~isempty(r.ind) % there are variables left in R

        % get variable F from R
        % with the maximum minimum association with T given EPC(T)        
        [f assocmin_f statMin_f tInEpc_f tInMepc_f r] = mmheur(r);

    end  
   
    
    if ~isempty(f) % there exist a variable F to add to EPC(T)

        % add F to EPC(T)
        Epc_t.ind = [ Epc_t.ind f ];
        % F's maximum p-value is F's negative minimum association
        Epc_t.pMax = [ Epc_t.pMax -assocmin_f ];
        Epc_t.pMaxStat = [ Epc_t.pMaxStat statMin_f ];
        Epc_t.tIndInEpc = [ Epc_t.tIndInEpc tInEpc_f ];
        Epc_t.tIndInMepc = [ Epc_t.tIndInMepc tInMepc_f ];

        if Params.verbose
            fprintf('\n%d entered EPC(%d).\n', f, t);
            fprintf('\nNew EPC(%d) = [ %s ].\n', t, num2str(Epc_t.ind));
        end

    end
    
    if isempty(Ppc_t.ind) && isempty(r.ind)    
        % no variable remained to consider adding to EPC(T)
        break;        
    end
    
end

end

function [Epc_t Nepc_t] = backward(subs, nLevels, t, Epc_t, Nepc_t, Params)
%PHASE2 Phase II (backward): remove variables from EPC(T)

if numel(Epc_t.ind) < 2
    % if 0, nothing to remove
    % if 1, there are no other variables in EPC(T) that could be affected
    % by the insertion of the last and only variable F
    return;
end

% initially all variables in EPC(T) are associated with T given EPC(T)
h = true(1, numel(Epc_t.ind));

for xInEpc_t = 1:numel(Epc_t.ind)-1 % for each variable in EPC(T) except the last one (F)
    
    % let it be X
    x = Epc_t.ind(xInEpc_t);
    
    % we exclude F, the last variable added to EPC(T),
    % because it is already known that F is associated with T
    % given any subset of the previous EPC(T)
        
    % determine the maximum size of a conditioning set S
    % that enables accurate calculation of I(X;T|S)
    nsmax_xt = findnsmax(subs, nLevels, x, t, Epc_t.ind(h), Params);
    
    if Params.verbose
        fprintf('\nmax |S| for a subset of EPC(%d)\\{%d} = [%s] in testing I(%d;%d|S): %d of %d\n', t, x, num2str(Epc_t.ind(h)), x, t, nsmax_xt, sum(h));
    end
    
    if nsmax_xt == 0
        continue;
    end
    
    % find indices of the rest variables in EPC(T) not yet removed;
    % in the worst case this in only the last variable (F)
    indices = setdiff( [ 1 : (xInEpc_t - 1) (xInEpc_t + 1) : numel(Epc_t.ind) ], find(~h) );
    
    if numel(indices) == 1
        
        % only I(X;T|F) is left to be tested
        
        % test I(X;T|F)
        f = Epc_t.ind(end);
        
        [h(xInEpc_t) p stat] = test(subs, nLevels,x, t, f, Params,  Epc_t.tIndInEpc(xInEpc_t), Epc_t.tIndInMepc(xInEpc_t));
        
        if ~h(xInEpc_t) || ( ~isnan(p) && pcmp(p, stat, Epc_t.pMax(xInEpc_t), Epc_t.pMaxStat(xInEpc_t)) > 0 )
            
            % update maximum p-value and its statistic
            % if independence was accepted or it was rejected
            % using a successfully computed (not NaN) p-value
            Epc_t.pMax(xInEpc_t) = p;
            Epc_t.pMaxStat(xInEpc_t) = stat;

        end
        
        break; % finished
        
    end
    
    for ns_xt = 1:nsmax_xt % for each size up to the maximum size
        
        % note: for X in EPC(T), I(X;T|[]) has been tested; ignoring size 0
        
        % find subsets of -indices- of variables in EPC(T) 
        % of this size that do not contain the index of X      
        C = nchoosek(indices, ns_xt);
        
        for k = 1:size(C, 1) % for each such index subset
            
            
            if max(C(k, :)) < xInEpc_t
                
                % subset is older in EPC(T) than X, test already performed
                continue;
                
            end
            
            % let it be S
            s = Epc_t.ind(C(k, :));

            % test I(X;T|S)
            [h(xInEpc_t) p stat] = test(subs, nLevels,x, t, s, Params,  Epc_t.tIndInEpc(xInEpc_t), Epc_t.tIndInMepc(xInEpc_t));                

            if ~h(xInEpc_t) || ( ~isnan(p) && pcmp(p, stat, Epc_t.pMax(xInEpc_t), Epc_t.pMaxStat(xInEpc_t)) > 0 )

                % update maximum p-value and its statistic
                % if independence was accepted or it was rejected
                % using a successfully computed (not NaN) p-value
                Epc_t.pMax(xInEpc_t) = p;
                Epc_t.pMaxStat(xInEpc_t) = stat;

            end
            
            if ~h(xInEpc_t)
                break;
            end
            
        end

        if ~h(xInEpc_t)
            break;
        end

    end

end
  
if Params.verbose
    fprintf('\nRemoving [ %s ] from EPC(%d).\n', num2str(Epc_t.ind(~h)), t);
end

isNotAssoc = ~h;

% add variables with no association with T given EPC(T) to NEPC(T)

Nepc_t.ind = [ Nepc_t.ind Epc_t.ind(isNotAssoc) ];

if Params.includeNepc
    
    Nepc_t.pMax = [ Nepc_t.pMax Epc_t.pMax(isNotAssoc) ];
    Nepc_t.pMaxStat = [ Nepc_t.pMaxStat Epc_t.pMaxStat(isNotAssoc) ];
    
end

% remove variables with no association with T given EPC(T) from EPC(T)
Epc_t.ind(isNotAssoc) = [];
Epc_t.pMax(isNotAssoc) = [];
Epc_t.pMaxStat(isNotAssoc) = [];
Epc_t.tIndInEpc(isNotAssoc) = [];      
Epc_t.tIndInMepc(isNotAssoc) = [];

end

function [z Nepc_t] = minassocupdate(subs, nLevels, t, Epc_t, Nepc_t, zname, z, Params)
% MINASSOCUPDATE update R or PPC(T)
% Calculates associations of the remaining variables in Z with T 
% given the new subsets of EPC(T), while removing any variables
% that achieve no association.

% initially all variables in Z are associated with T given EPC(T)
h = true(1, numel(z.ind));        

for xIndInZ = 1:length(z.ind) % for each variable in Z

    % let it be X
    x = z.ind(xIndInZ);
        
    if isempty(Epc_t.ind)

        % test I(X;T|[])
        [h(xIndInZ) p stat] = test(subs, nLevels,x, t, [], Params,  z.tIndInEpc(xIndInZ), z.tIndInMepc(xIndInZ));

        assoc = -p;

        if ~h(xIndInZ) || (~isnan(assoc) && assoccmp(assoc, stat, z.assocMin(xIndInZ), z.assocMinStat(xIndInZ)) < 0)

            % update minimum association and its statistic
            % if independence was accepted or it was rejected
            % using a successfully computed (not NaN) p-value
            z.assocMin(xIndInZ) = assoc;
            z.assocMinStat(xIndInZ) = stat;

        end
        
        continue;

    end

    % determine the maximum size of a conditioning set S
    % that enables accurate testing of I(X;T|S)
    nsmax_xt = findnsmax(subs, nLevels, x, t, Epc_t.ind, Params);

    if Params.verbose
        fprintf('\nmax |S| for a subset of EPC(%d) = [%s] in testing I(%d;%d|S): %d of %d\n', t, num2str(Epc_t.ind), x, t, nsmax_xt, numel(Epc_t.ind));
    end

    if nsmax_xt == 0 
        
        % only I(X;T|[]) can be tested accuratelly
        % and it's already tested
        continue;
        
    end % else nsmax_xt > 0

    % test I(X;T|F)
    f = Epc_t.ind(end);
    [h(xIndInZ) p stat] = test(subs, nLevels,x, t, f, Params,  z.tIndInEpc(xIndInZ), z.tIndInMepc(xIndInZ));            

    assoc = -p;
        
    if ~h(xIndInZ) || (~isnan(assoc) && assoccmp(assoc, stat, z.assocMin(xIndInZ), z.assocMinStat(xIndInZ)) < 0)

        % update minimum association and its statistic
        % if independence was accepted or it was rejected
        % using a successfully computed (not NaN) p-value
        z.assocMin(xIndInZ) = assoc;
        z.assocMinStat(xIndInZ) = stat;

    end

    if ~h(xIndInZ)
        continue; % no need to try more subsets
    end

    if nsmax_xt < 2
        % |S| == 0 ([]) and |S| == 1 (F) already tested,
        % no other |S| < 2 to test
        continue;                
    end

    if numel(Epc_t.ind(end)) == 2 % if |EPC(T)| == 2
    
        % test I(X;T|EPC(T))
        [h(xIndInZ) p stat] = test(subs, nLevels,x, s, Epc_t.ind, Params,  z.tIndInEpc(xIndInZ), z.tIndInMepc(xIndInZ));            

        assoc = -p;

        if ~h(xIndInZ) || (~isnan(assoc) && assoccmp(assoc, stat, z.assocMin(xIndInZ), z.assocMinStat(xIndInZ)) < 0)

            % update minimum association and its statistic
            % if independence was accepted or it was rejected
            % using a successfully computed (not NaN) p-value
            z.assocMin(xIndInZ) = assoc;
            z.assocMinStat(xIndInZ) = stat;

        end

        % we tested with all subsets of both variables,
        % no more subsets to try
        continue;
    
    end % else |EPC(T)| > 2

    for ns_xt = 2:nsmax_xt % for each size up to the maximum size

        % find subsets of this size - 1 among the previous EPC(T)
        C = nchoosek( Epc_t.ind(1:end-1), ns_xt - 1); 

        for k = 1:size(C, 1) % for each such subset

            % add F and let it be S 
            s = [ C(k, :) Epc_t.ind(end) ];

            % test I(X;T|S)
            [h(xIndInZ) p stat] = test(subs, nLevels,x, t, s, Params,  z.tIndInEpc(xIndInZ), z.tIndInMepc(xIndInZ));                

            assoc = -p;

            if ~h(xIndInZ) || (~isnan(assoc) && assoccmp(assoc, stat, z.assocMin(xIndInZ), z.assocMinStat(xIndInZ)) < 0)

                % update minimum association and its statistic
                % if independence was accepted or it was rejected
                % using a successfully computed (not NaN) p-value
                z.assocMin(xIndInZ) = assoc;
                z.assocMinStat(xIndInZ) = stat;

            end

            if ~h(xIndInZ)
                break;
            end

        end

        if ~h(xIndInZ)
            break;
        end
        
    end            
    
end

if Params.verbose
    fprintf('\nRemoving [ %s ] from %s.\n', num2str(z.ind(~h)), zname);
    fprintf('\nNew %s = [ %s ].\n', zname, num2str(z.ind(h)));
end

isNotAssoc = ~h;

% add variables with no association with T given EPC(T) to NEPC(T)

Nepc_t.ind = [ Nepc_t.ind z.ind(isNotAssoc) ];

if Params.includeNepc
    
    Nepc_t.pMax = [ Nepc_t.pMax -z.assocMin(isNotAssoc) ];
    Nepc_t.pMaxStat = [ Nepc_t.pMaxStat z.assocMinStat(isNotAssoc) ];

end

% remove variables with no association with T given EPC(T) from Z
z.ind(isNotAssoc) = [];
z.assocMin(isNotAssoc) = [];
z.assocMinStat(isNotAssoc) = [];
z.tIndInEpc(isNotAssoc) = [];
z.tIndInMepc(isNotAssoc) = [];

end

function [h p stat] = test(subs, nLevels,x, t, s, Params, tIndInEpc_x, tIndInMen_x)
% TEST
  
if Params.verbose
    fprintf('\nFinding I(%d; %d | [ %s ])...', x, t, num2str(s));
end

h = NaN; % initially we don't know

if ~isnan(tIndInEpc_x)

    % if T is in in EPC(X)

    if all(ismember(s, Params.Epc.ind{x})) % if S is a subset of EPC(X) too
        
        % note: [] is a subset of all sets

        % we have already performed this test
        % ({T} U S) subset of EPC(X) => ~I(X;T|S)
        h = 1;

        % debug
        %fprintf('\nReusing from EPC(%d)...\n', x);
        %[h_true p_true] = Params.testFun(subs, nLevels, x, t, s, Params.alpha, Params.omin);
        %assert( isnan(p_true) || h == 1);

        % get the maximum p-value
        % TODO: explain why valid
        p = Params.Epc.pMax{x}(tIndInEpc_x);
        stat = Params.Epc.pMaxStat{x}(tIndInEpc_x);

        if Params.verbose
            fprintf(' Reusing from EPC(%d)...', x);
        end              

    elseif ~isnan(tIndInMen_x) && all(ismember(s, Params.Mepc.ind{x})) % if S is in MEPC(X)

        % find indices of the variables of S in MEPC(X)
        sIndInMepc_x = zeros(size(s));
        for i=1:numel(s)
            sIndInMepc_x(i) = find(Params.Mepc.ind{x} == s(i), 1);
        end

        if isempty(s) || max(sIndInMepc_x) < tIndInMen_x

            % if S was a subset of EPC(X) before T entered
            % ([] is always such a subset)
            % we have already calculated this association
            % {T} U S has been a subset of EPC(X)
            % => ~I(X;T|S) <=> Assoc(X;T|S)
            h = 1;

            % debug
            %fprintf('\nReusing from MEPC(%d)...\n', x);
            %[h_true p_true] = Params.testFun(subs, nLevels, x, t, s, Params.alpha, Params.omin);
            %assert( isnan(p_true) || h == 1);

            % get the maximum p-value
            % TODO: explain why valid
            p = Params.Epc.pMax{x}(tIndInEpc_x);
            stat = Params.Epc.pMaxStat{x}(tIndInEpc_x);

            if Params.verbose
                fprintf(' Reusing from MEPC(%d)...', x);
            end

        end

    end

end

if isnan(h) % not tested yet or no reuse

    if Params.verbose
        fprintf(' Computing...');
    end

    % test it
    [h p stat] = Params.testFun(subs, nLevels, x, t, s, Params.alpha, Params.omin);
    
    % if the test wasn't performed and s isn't empty
    % assume dependence, not independence
    if isnan(p) && ~isempty(s)
        h = 1;
    end

end

if Params.verbose
    if h == 0
        fprintf(' Yes (%f. %f).\n', p , stat);
    else
        fprintf(' No (%f, %f).\n', p , stat);
    end
end

end

function [f assocmin_f statMin_f tInEpc_f tInMepc_f Ppc_t] = ppcnext(Ppc_t)
%PNNEXT get next variable from PPC(T) 
% Gets next variable from PPC(T) assuming that there are (associated) 
% variables left in PPC(T).

% note: rl >= 2

% get F
f = Ppc_t.ind(1);    
assocmin_f = Ppc_t.assocMin(1);
statMin_f = Ppc_t.assocMinStat(1);
tInEpc_f = Ppc_t.tIndInEpc(1);
tInMepc_f = Ppc_t.tIndInMepc(1);  

% remove F from R
Ppc_t.ind(1) = [];
Ppc_t.assocMin(1) = [];
Ppc_t.assocMinStat(1) = [];
Ppc_t.tIndInEpc(1) = [];
Ppc_t.tIndInMepc(1) = [];

end

function [f assocmin_f statMin_f tInEpc_f tInMepc_f r] = mmheur(r)
% MMHEUR Max-min heuristic

% assuming there are variables (with association) left in R

% get the index (among the remaining variables) and minimum association
% of the variable with the maximum minimum association
% among the remaining variables

assocminmax = -1;
statassocminmax = 0;

for xIndInR = 1:numel(r.ind) % for each variable in R
    
    if assoccmp(r.assocMin(xIndInR), r.assocMinStat(xIndInR), assocminmax, statassocminmax) > 0
        
        fIndInR = xIndInR;
        assocminmax = r.assocMin(xIndInR);
        statassocminmax = r.assocMinStat(xIndInR);
        
    end        
    
end

% get F
f = r.ind(fIndInR);    
assocmin_f = r.assocMin(fIndInR);
statMin_f = r.assocMinStat(fIndInR);   
tInEpc_f = r.tIndInEpc(fIndInR);
tInMepc_f = r.tIndInMepc(fIndInR);

% remove F from R
r.ind(fIndInR) = [];
r.assocMin(fIndInR) = [];
r.assocMinStat(fIndInR) = [];
r.tIndInEpc(fIndInR) = [];
r.tIndInMepc(fIndInR) = [];

end

function k = findnsmax(subs, nLevels, x, t, smax, Params)
% determine the maximum size of a conditioning set S that enables accurate
% calculation of I(X;T|S)

nobs = size(subs, 1);

nLevelsMin = min(nLevels([x t smax]));

k = round(log(nobs/(Params.omin*nLevels(t)*nLevelsMin))/log(nLevelsMin));
k = max(min([k Params.nsmax]), 0);

k = min(k, numel(smax));

%{
%fprintf('\n%d of %d\n', k, numel(Epc_t.ind));

if k < numel(Epc_t.ind)
    disp('Limited conditioning set.');
end
%}

end

function pcmp = pcmp(p1, stat1, p2, stat2)
% returns -1, 0, or 1 if the first p-value
% is less than, equal to, or greater than the second one.

if p1 < p2
    pcmp = -1; 
elseif p1 > p2
    pcmp = 1; 
else
    
    if stat1 > stat2
        pcmp = -1;
    elseif stat1 < stat2
        pcmp = 1;
    else
        pcmp = 0;
    end
    
end
            
end

function assoccmp = assoccmp(assoc, stat, assocMin, pMaxStat)

assoccmp = pcmp(assoc, pMaxStat, assocMin , stat);
            
end  