function [skeleton pMax pMaxStat] = mmskeleton(a, varargin)
%MMSKELETON learn skeleton of a bayesian network using MMPC-bar algorithm
% 
%   SKELETON = MMSKELETON(A) learns the skeleton of a bayesian network from
%   the samples of its variables in dataset A by applying the MMPC-bar
%   algorithm to each of the variables and performing hypothesis tests of 
%   conditional independence at the (100*(ALPHA=0.05))% significance level.
%   The learned skeleton is returned in the M-by-M sparse adjacency matrix
%   SKELETON, where M is the number of variables in dataset A. Nonzero 
%   entries in SKELETON indicate the presence of an edge. A must be a
%   dataset with categorical only variables. The network is assumed to be
%   ALPHA-faithful, i.e. two variables are adjacent iff all the p-values of
%   the tests of their conditional independence given any subset of the
%   other variables are <= ALPHA. When this assumption holds, no false 
%   negatives are returned (no true edges are missing), and ALPHA is the 
%   expected false positive rate, i.e. the expected proportion of non-edges
%   returned by the algorithm as edges.
%
%   [SKELETON, PMAX] = MMSKELETON(A) also returns a sparse M-by-M matrix 
%   PMAX containing upper bounds on the p-values of the complex hypothesis
%   of existence of the edges specified in SKELETON. These bounds are the 
%   maximum p-values among the tests of the primitive hypotheses of
%   conditional independence between edge ends. Elements of PMAX
%   corresponding to non-edges are zeros.
%
%   [SKELETON, PMAX, PMAXSTAT] = MMSKELETON(A) also returns a sparse M-by-M
%   matrix PMAXSTAT containing the G-square statistics corresponding to the
%   p-values in PMAX. Elements of PMAXSTAT corresponding to non-edges are
%   zeros.
%
%   [...] = MMSKELETON(A, 'Param1',VAL1,'Param2',VAL2,...) specifies
%   additional parameter name/value pairs chosen from the following:
%
%      'alpha'          A value between 0 and 1 that specifies the
%                       significance level as (100*ALPHA)% for the
%                       hypothesis tests of conditional independence that
%                       are performed. Default is 0.05.
%
%      'omin'           A non-negative integer that specifies the minimum
%                       average observed count per cell (bin) required for
%                       a test to be performed. MMSKELETON does not
%                       perform a test if the average observed count per 
%                       cell is smaller than OMIN and accepts the null 
%                       hypothesis if the conditioning set is the empty set
%                       otherwise it rejects it. Default is 5.
%
%      'nsmax'          A non-negative integer that specifies the maximum
%                       cardinality of a conditioning set. MMSKELETON
%                       does not perform hypothesis tests of conditional
%                       independence conditioning on sets with cardinality
%                       greater than NSMAX. Default is 10.
%
%      'testFun'        A handle to a function that performs a hypothesis
%                       test of conditional independence. Default is
%                       ORG.MENSXMACHINA.COMMONS.STAT.CHI2TESTOPT, which
%                       performs a G-square test of conditional
%                       independence. See the documentation of that
%                       function for requirements of a conditional
%                       independence test function.
%
%      'sc'             A logical scalar that specifies whether symmetry
%                       correction will be applied or not. If logical 1
%                       (true), two variables X and T are considered 
%                       adjacent iff X is in EPC(T) and T is in EPC(X), 
%                       otherwise iff X is in EPC(T) or T is in EPC(X). In
%                       the latter case, the PMAX and PMAXSTAT matrices
%                       returned are not symmetric: PMAX(T,X) contains the
%                       maximum p-value among the hypothesis tests of 
%                       conditional independence between T and X made when
%                       learning EPC(T), while PMAX(X,T) contains the
%                       maximum p-value among the hypothesis tests of
%                       conditional independence between T and X made when
%                       learning EPC(X); PMAXSTAT contains the 
%                       corresponding statistics. DefaultSC is true.
%
%      'rl'             A value between 0 and 3 specifying the level of 
%                       reuse RL that takes place between subsequent calls
%                       of MMPC. If RL is 0, no reuse takes place and the 
%                       same hypothesis tests are performed again and again
%                       in each subsequent call. If RL is 1, each call uses
%                       the NEPC (Non-EPC) sets returned by the previous
%                       calls in order to reuse hypothesis test results 
%                       implied by that sets: If T is is NEPC(X) (i.e. not
%                       in EPC(X)), then it is known that T and X are 
%                       conditionally independent given some subset. RL 1
%                       has been found to result in faster execution times 
%                       than any other RL in our tests. If RL is 2, each 
%                       call also uses the EPC sets: if T and a subset S 
%                       are in EPC(X), then it is known that X and T are 
%                       conditionally dependent given S. Finally, if RL is
%                       3, each call also uses the EPC sets returned in the
%                       first (forward) phase, called "maximal" EPC
%                       ("MEPC") of the previous calls in order to reuse
%                       hypothesis test results implied by that sets: if T 
%                       and a subset S are in MEPC(X), then it is known
%                       that X and T are conditionally dependent given S. 
%                       All the hypothesis test results of the previous 
%                       calls of MMPC are implied by the EPC and MEPC sets,
%                       so RL 3 achieves complete reuse. However complete 
%                       reuse does not usually translate in faster 
%                       execution times (see tests). RL 3 and 2 always
%                       produce the  same skeleton, but RL 0 and 1 may
%                       produce different ones. Default RL is 1.
%
%   'includeNonEdges'   A logical scalar that specifies whether data about 
%                       both  edges and non-edges are returned. If logical 
%                       1 (true), PMAX and PMAXSTAT are full matrices that
%                       contain the maximum p-values and their
%                       corresponding G-square statistics respectively for
%                       all pairs of variables in A whether they are 
%                       neighbors or not. A maximum p-value corresponding
%                       to a non-edge is the p-value of the last testing of
%                       conditional independence of its ends, which
%                       determined its status as a non-edge, and it is
%                       either > ALPHA or NaN, being the latter in the case
%                       that the test of conditional independence of the
%                       two ends given the empty set wasn't performed and
%                       subsequently independence was accepted. Default is
%                       false.
%
%      'verbose'        A logical scalar that specifies whether messages
%                       are displayed during the execution of the
%                       function. Default is false.
%
%   Example:
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % learn ALARM's skeleton
%       alarm_skeleton = org.mensxmachina.bnet.mmskeleton(a);
%
%   See also ORG.MENSXMACHINA.STAT.CHI2TESTOPT,
%       ORG.MENSXMACHINA.BNET.MMPC, ORG.MENSXMACHINA.BNET.MMPCBAR.

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

import org.mensxmachina.bnet.*;
import org.mensxmachina.commons.stats.*;

p = inputParser;  % Create an instance of the inputParser class
p.FunctionName = 'ORG.MENSXMACHINA.BNET.MMSKELETON';

p.addRequired('a', @iscategoricaldataset);

p.addParamValue('alpha', 0.05, @(alpha) isscalar(alpha) && alpha >= 0 && alpha <= 1);
p.addParamValue('omin', 5, @(omin) isscalar(omin) && omin>=0 && isreal(omin) && omin==round(omin));
p.addParamValue('nsmax', 10, @(nsmax) isscalar(nsmax) && nsmax>=0 && isreal(nsmax) && nsmax==round(nsmax));
p.addParamValue('testFun', @chi2testopt, @(testFun) isa(testFun, 'function_handle'));
p.addParamValue('sc', true, @(sc) isscalar(sc) && islogical(sc));
p.addParamValue('rl', 1, @(rl) isreal(rl) && isscalar(rl) && ismember(rl, [0 1 2 3]));
p.addParamValue('includeNonEdges', false, @islogical);
p.addParamValue('verbose', false, @islogical);
p.addParamValue('subs', [], @(subs) isnumeric(subs) && ndims(subs) == 2 );
p.addParamValue('nLevels', [], @(nLevels) isnlevels(a, nLevels) );

p.parse(a, varargin{:});

Params = p.Results;

subs = p.Results.subs;
nLevels = p.Results.nLevels;

Params = rmfield(Params, {'subs', 'nLevels'});

if isempty(subs)
    % convert dataset to subscripts
    subs = double(a);
end 

if isempty(nLevels)
    % get # of levels
    nLevels = nlevels(a);
end

% get V, the set of all variables
v = 1:size(a, 2);

nv = numel(v);

% initialize EPC and NEPC, the sets of EPC(T) and NEPC(T) for each T in V
if Params.rl >= 1
    
    Params.Nepc.ind = cell(1, nv);
    Params.Nepc.pMax = cell(1, nv);
    Params.Nepc.pMaxStat = cell(1, nv);
    
    if Params.rl >= 2
    
        Params.Epc.ind = cell(1, nv);
        Params.Epc.pMax = cell(1, nv);
        Params.Epc.pMaxStat = cell(1, nv);

        % initialize MEPC, the set of MEPC(T) for each T in V
        if Params.rl == 3
            Params.Mepc.ind = cell(1, nv);    
        else
            Params.Mepc = []; 
        end
    
    else
        Params.Epc = [];
        Params.Mepc = []; 
    end
    
else
    Params.Nepc = [];
    Params.Epc = [];    
    Params.Mepc = [];
end

% initialize skeleton
skeleton = sparse(nv, nv);

% initialize p-value and statistic matrices
if Params.includeNonEdges
    pMax = zeros(nv, nv);
    pMaxStat = zeros(nv, nv);
else
    pMax = sparse(nv, nv);
    pMaxStat = sparse(nv, nv);   
end

for t = v % for each variable T in V
    
    % learn EPC(T)
    fprintf('\nLearning EPC(%d)...\n', t);
    
    Params.includeNepc = Params.includeNonEdges;
    
    % call optimized version of MMPC-bar
    [epc_t_ind pMax_t pMaxStat_t Params.Nepc Params.Epc Params.Mepc] = ...
        org.mensxmachina.bnet.mmpcbaropt(subs, nLevels, t, Params);
    
    if Params.includeNonEdges
        
        skeleton(t, :) = epc_t_ind;
        pMax(t, :) = pMax_t;
        pMaxStat(t, :) = pMaxStat_t;
        
    elseif ~isempty(epc_t_ind)
        
        skeleton(t, epc_t_ind) = 1;
        pMax(t, epc_t_ind) = pMax_t;
        pMaxStat(t, epc_t_ind) = pMaxStat_t;
        
    end
    
end

if Params.sc % symmetry correction on

    % consider Adj(X,T) <=> X in EPC(T) AND T in EPC(X)
    skeleton = tril(skeleton) & triu(skeleton)';

    % make skeleton symmetric
    skeleton = skeleton + skeleton';
    
    % get the lower triangular p-value matrix
    pMax_tril = tril(pMax, -1);
    
    % get the transposed lower triangular p-value matrix
    pMax_triu = triu(pMax, 1)';
    
    % find which p-values in the transposed upper triangular matrix
    % are greater than the respective ones in the lower triangular
    isMaxPMax_triu = pMax_triu > pMax_tril;
    
    % replace these p-values in the lower triangular;
    % the diagonal will stay as it is (all ones)
    pMax_tril(isMaxPMax_triu) = pMax_triu(isMaxPMax_triu);
    
    % make p-value matrix symmetric and set diagonal to all ones
    if Params.includeNonEdges
        pMax = pMax_tril + pMax_tril' + eye(nv, nv);    
    else
        pMax = pMax_tril + pMax_tril' + sparse(eye(nv, nv)); 
    end
    
    % get the lower triangular statistics matrix
    pMaxStat_tril = tril(pMaxStat, -1);
    
    % get the transposed lower triangular statistics matrix
    pMaxStat_triu = triu(pMaxStat, 1)';
    
    % replace statistics in the lower triangular with the corresponding
    % statistics in the upper triangular that correspond to the max p-value
    pMaxStat_tril(isMaxPMax_triu) = pMaxStat_triu(isMaxPMax_triu);
    
    % make statistic matrix symmetric
    pMaxStat = pMaxStat_tril + pMaxStat_tril';
    
else % symmetry correction off

    % consider Adj(X,T) <=> X in EPC(T) OR T in EPC(X)
    skeleton = tril(skeleton) | triu(skeleton)'; 

    % make skeleton symmetric
    skeleton = skeleton + skeleton';
    
end

end