function [auc tpr fpr threshold] = pcauc(targetskeleton, pMax, t)
%PCAUC Parents and children learner area under the ROC curve (AUC)
%
%   PCAUC is called like ORG.MENSXMACHINA.BNET.PCROC, but it also 
%   calculates the area(s) under the ROC curve for each of the inputs
%   and returns them in row vector AUC, as the first output argument.
%
%   Examples:
%
%       % load ALARM network
%       load mxm_bnet_alarm_net; % loads 'net' Network
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % learn parents and children set of variable #37
%       [pc pMax pMaxStat] = ...
%           org.mensxmachina.bnet.mmpc(a, 37, 'includeNpc', true);
%
%       % get true skeleton
%       skeleton_true = net.skeleton;
%
%       % calculate ROC points and AUC for variable #37
%       [auc tpr fpr threshold] = ..
%           org.mensxmachina.bnet.pcauc(skeleton_true, pMax, 37);
%
%   See also: 
%       ORG.MENSXMACHINA.COMMONS.STATS.ROC,
%       ORG.MENSXMACHINA.BNET.PLOTPCROC,
%       ORG.MENSXMACHINA.BNET.PCAUC 

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

if ~iscell(targetskeleton)

    % get roc inputs
    [targets outputs] = rocinputs(targetskeleton, pMax, t);

else
        
    targets = cell(size(targetskeleton));
    outputs = cell(size(targetskeleton));
    
    for i=1:length(targetskeleton)
    
        % get roc inputs
        [targets{i} outputs{i}] = ...
            rocinputs(targetskeleton{i}, pMax{i}, t(i));
    
    end
    
end

% calculate AUC
[auc tpr fpr threshold] = ...
    org.mensxmachina.commons.stats.auc(targets, outputs);

end

function [targets outputs] = rocinputs(targetskeleton, pMax, t)

targets = targetskeleton(t, :);
targets(t) = []; % remove target variable

outputs = -pMax;
outputs(t) = []; % remove target variable

end