function cp = pcperf(targetskeleton, output, t)
%PCPERF Evaluate performance of parents and children learner(s)
%
%   ORG.MENSXMACHINA.BNET.PCPERF is a version of the CLASSPERF function in
%   the Bioinformatics Toolbox (TM), specialized for parents and children
%   learning. For a given node, the task of learning its parents and
%   children can be viewed as a binary classification task, where each
%   other node is classified as either parent or child, or not.
%
%   CP = PCPERF(TARGETSKELETON, OUTPUTPC, T) returns a classifier
%   performance object CP for the output parents and children set OUTPUTPC,
%   returned either by ORG.MENSXMACHINA.BNET.MMPC or
%   ORG.MENSXMACHINA.BNET.MMPCBAR, targeting the parents and children of
%   the variable specified by positive integer T in skeleton
%   TARGETSKELETON.
%
%   CPS = PCPERF(TARGETSKELETONS, OUTPUTPCS, TS), where TARGETSKELETONS is
%   1-by-N (1-by-N) cell array of skeletons, OUTPUTPCS is a 1-by-N (1-by-N)
%   cell array of output parents and children sets and TS is a 1-by-N
%   (1-by-N) positive integer vector, creates a classifier performance
%   object for each of the parents and children learning tasks
%   corresponding to TARGETSKELETONS, OUTPUTPCS and TS triplets and returns
%   them in 1-by-N (1-by-N) cell array of classifier performance objects
%   CPS.
%
%   Examples:
%
%       % load ALARM network
%       load mxm_bnet_alarm_net; % loads 'net' Network
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % learn parents and children set of variable #37
%       pc = org.mensxmachina.bnet.mmpc(a, 37);
%
%       % get true skeleton
%       skeleton_true = net.skeleton;
%
%       % get classifier performance object for variable #37
%       cp = org.mensxmachina.bnet.pcperf(skeleton_true, pc, 37);
%
%   See also: CLASSPERF, ORG.MENSXMACHINA.BNET.SKELETONPERF

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

if iscell(targetskeleton)

    cp = cell(size(targetskeleton));
    
    for i=1:length(targetskeleton) % for each triplet

        % get parents and children learning performance
        cp{i} = pcperf_single(targetskeleton{i}, output{i}, t(i));

    end

else
        
    % get parents and children learning performance
    cp = pcperf_single(targetskeleton, output, t);
    
end

end

function cp = pcperf_single(targetskeleton, outputpc, t)

targets = targetskeleton(t, :);
targets(t) = []; % remove target variable

if issparse(outputpc)
    outputs = outputpc;  
else   
    % convert to logical indices
    outputs = false(1, size(targetskeleton, 1));
    outputs(outputpc) = true;    
end

outputs(t) = []; % remove target variable

% get parents and children learning performance
cp = classperf(double(targets), double(outputs), 'Positive', 1, 'Negative', 0); 

end