function [tpr fpr threshold] = pcroc(targetskeleton, pMax, t)
%PCROC Parents and children learning receiver operating characteristic
%
%   PCROC is a version of the ORG.MENSXMACHINA.COMMONS.STATS.ROC
%   function specialized for parents and children learning. For a given
%   node, the task of learning its parents and children can be viewed as a
%   binary classification task, where each other node is classified as
%   either parent or child, or not.
%
%   [TPR,FPR,THRESHOLD] = PCROC(TARGETSKELETON, PMAX, T), where PMAX is the
%   maximum p-values vector returned by either ORG.MENSXMACHINA.BNET.MMPC
%   (with 'includeNpc' parameter set to true) or
%   ORG.MENSXMACHINA.BNET.MMPCBAR (with 'includeNepc' parameter set to
%   true) targeting the variable in adjacency matrix TARGETSKELETON 
%   specified by the positive integer T, calculates for each threshold
%   (here the threshold are the negative maximum p-values) the true 
%   positive and false positive rate of the parents and children learner 
%   and returns them in row vectors TPR and FPR, while the corresponding
%   threshold are returned in the THRESHOLD row vector. Each pair of 
%   values (TPR(k), FPR(k)) defines a point in the Receiver operating
%   characteristic (ROC) curve.
%
%   [TPRS,FPRS,THRESHOLDS] = PCROC(TARGETSKELETONS, PMAXS, TS), where 
%   TARGETSKELETONS is a 1-by-N (1-by-N) cell array of skeletons, PMAXS is
%   a 1-by-N (1-by-N) array of maximum p-value vectors and TS is a 1-by-N
%   (1-by-N) positive integer vector, calculates rates and threshold for 
%   each of the parents and children learning tasks corresponding to
%   TARGETSKELETONS, OUTPUTPCS and TS triplets and returns them in the
%   1-by-N (1-by-N) cell arrays of row vectors TPRS and FPRS respectivelly,
%   while the corresponding threshold are returned in the 1-by-N (1-by-N)
%   cell array of row vectors THRESHOLDS.
%
%   Example:
%
%       % load ALARM network
%       load mxm_bnet_alarm_net; % loads 'net' Network
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % learn parents and children set of variable #37
%       [pc pMax pMaxStat] = ...
%           org.mensxmachina.bnet.mmpc(a, 37, 'includeNpc', true);
%
%       % get true skeleton
%       skeleton_true = net.skeleton;
%
%       % calculate ROC for variable #37
%       [tpr fpr threshold] = ..
%           org.mensxmachina.bnet.pcroc(skeleton_true, pMax, 37);
%
%   See also: 
%       ORG.MENSXMACHINA.COMMONS.STATS.ROC,
%       ORG.MENSXMACHINA.BNET.PLOTPCROC,
%       ORG.MENSXMACHINA.BNET.PCAUC 

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

if ~iscell(targetskeleton)

    % get roc inputs
    [targets outputs] = rocinputs(targetskeleton, pMax, t);

else
        
    targets = cell(size(targetskeleton));
    outputs = cell(size(targetskeleton));
    
    for i=1:length(targetskeleton)
    
        % get roc inputs
        [targets{i} outputs{i}] = ...
            rocinputs(targetskeleton{i}, pMax{i}, t(i));
    
    end
    
end

% calculate ROC
[tpr fpr threshold] = ...
    org.mensxmachina.commons.stats.roc(targets, outputs);

end

function [targets outputs] = rocinputs(targetskeleton, pMax, t)

targets = targetskeleton(t, :);
targets(t) = []; % remove target variable

outputs = -pMax;
outputs(t) = []; % remove target variable

end