function [tpr fpr thresholds] = skeletonroc(targetskeleton, pMax)
%SKELETONROC Skeleton learner receiver operating characteristic
%
%   SKELETONROC is a version of the ORG.MENSXMACHINA.COMMONS.STATS.ROC
%   function specialized for skeleton learning. A skeleton learning task
%   can be viewed as a binary classification task, where the possible edges
%   are classified as edges (positives) or non-edges (negatives).
%
%   [TPR,FPR,THRESHOLD] = SKELETONROC(TARGETSKELETON, PMAX) where 
%   TARGETSKELETON is an adjacency matrix and PMAX is the matrix of maximum
%   p-values, of the same size as TARGETSKELETON, as returned by
%   ORG.MENSXMACHINA.BNET.MMSKELETON targeting TARGETSKELETON and with
%   parameter 'includeNonEdges' set to true, calculates for each threshold
%   (here the thresholds are the negative maximum p-values) the true 
%   positive and false positive rate of the skeleton learner and returns
%   them in row vectors TPR and FPR, while the corresponding thresholds 
%   are returned in the THRESHOLD row vector. Each pair of values (TPR(k),
%   FPR(k)) defines a point in the Receiver operating characteristic (ROC)
%   curve.
%
%   [TPRS,FPRS,THRESHOLDS] = SKELETONROC(TARGETSKELETONS, PMAXS) where
%   TARGETSKELETONS is a 1-by-N (N-by-1) cell array of skeletons and PMAXS
%   is a same-sized cell of maximum p-value matrices, calculates rates and
%   thresholds for each TARGETSKELETONS and PMAXS pair and returns them 
%   in the 1-by-N (N-by-1) cell arrays of row vectors TPRS and FPRS 
%   respectivelly, while the corresponding thresholds are returned in the
%   1-by-N (N-by-1) cell array of row vectors THRESHOLDS. 
%
%   Example:
%
%       % load ALARM network samples
%       load mxm_bnet_alarm_samples; % loads 'a' dataset
%
%       % learn skeleton
%       [skeleton pMax pMaxStat] = ...
%           org.mensxmachina.bnet.mmskeleton(a, 'includeNonEdges', true);
%
%       % load ALARM network
%       load mxm_bnet_alarm_net; % loads 'net' network
%
%       % get true skeleton
%       targetskeleton = net.skeleton;
%
%       % calculate ROC
%       [tpr fpr threshold] = ...
%           org.mensxmachina.bnet.skeletonroc(targetskeleton, pMax)
%
%   See also: 
%       ORG.MENSXMACHINA.COMMONS.STATS.PROC,
%       ORG.MENSXMACHINA.BNET.PLOTSKELETONROC,
%       ORG.MENSXMACHINA.BNET.SKELETONAUC 

%   Copyright 2010 Mens X Machina
%
%   This file is part of Mens X Machina Bayesian Network Toolbox.
%
%   Mens X Machina Bayesian Network Toolbox is free software: you can
%   redistribute it and/or modify it under the terms of the GNU General
%   Public License as published by the Free Software Foundation, either 
%   version 3 of the License, or (at your option) any later version.
%
%   Mens X Machina Bayesian Network Toolbox is distributed in the hope that
%   it will be useful, but WITHOUT ANY WARRANTY; without even the implied
%   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
%   the GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with Mens X Machina Bayesian Network Toolbox. If not, see
%   <http://www.gnu.org/licenses/>.

% convert skeletons to vectors
targets = org.mensxmachina.commons.graph.skeleton2vector(targetskeleton);

% convert p-value matrices to vectors
outputs = org.mensxmachina.commons.graph.skeleton2vector(pMax);

if ~iscell(pMax)
    outputs = -outputs;
else
    outputs = cellfun(@uminus, outputs, 'UniformOutput', false);
end

[tpr fpr thresholds] = ...
    org.mensxmachina.commons.stats.roc(targets, outputs);

end