classdef TestMmpc < TestCase
%TestMmpc mmpc test cases

properties
    
    net % net info
    
    netParams % net parameters
    sampleParams % sampling parameters
    learnParams % learning parameters
    
    config_default
    
end

methods

function obj = TestMmpc(name)

obj = obj@TestCase(name);

%netsave();
%netssave();
%netparamssave();
%sampleparamssave();
%samplessave();

% load net info
load net;
obj.net = net;

% load net parameters
load netparams;
obj.netParams = netParams;
obj.netParams{'netName', 'values'} = obj.netParams{'netName', 'values'}(1:2);

% load sample parameters
load sampleparams;
obj.sampleParams = sampleParams;

obj.learnParams = dataset( ...
    {{[0 1 2 3]'; [0.01 0.05 0.1]'; [ true; false ]; [ true; false ]}, 'values'}, ...
    {{1; 0.05; false; false}, 'default'}, ...
    'ObsNames', {'rl', 'alpha', 'includeNpc', 'verbose'} ...
    );

% create default config
obj.config_default = dataset( ...
    {{obj.netParams{'netName', 'default'}}, 'netName'}, ...
    {obj.netParams{'targetVar', 'default'}, 'targetVar'}, ...
    {obj.netParams{'nsmax', 'default'}, 'nsmax'}, ...
    {obj.sampleParams{'sampleNo', 'default'}, 'sampleNo'}, ...
    {obj.sampleParams{'sampleSize', 'default'}, 'sampleSize'}, ...
    {obj.learnParams{'rl', 'default'}, 'rl'}, ...
    {obj.learnParams{'alpha', 'default'}, 'alpha'}, ...
    {obj.learnParams{'includeNpc', 'default'}, 'includeNpc'}, ...
    {obj.learnParams{'verbose', 'default'}, 'verbose'} ...
    );

end

function testIncludeNpc(obj)

Config = repmat(obj.config_default, length(obj.learnParams{'includeNpc', 'values'}), 1);

Config.includeNpc = obj.learnParams{'includeNpc', 'values'};

Run = obj.configRun(Config);

save('mmpc_run_var_includenpc', 'Run');

pcInd = find(Run.pc{1});

assertEqual( pcInd, Run.pc{2} );

assertTrue( issparse(Run.pc{1}) );
assertTrue( ~issparse(Run.pMax{1}) );
assertTrue( ~issparse(Run.pMaxStat{1}) );

assertEqual( Run.pMax{1}(pcInd), Run.pMax{2} );
assertEqual( Run.pMaxStat{1}(pcInd), Run.pMaxStat{2} );

end

function testVerbose(obj)

Config = repmat(obj.config_default, length(obj.learnParams{'verbose', 'values'}), 1);
Config.verbose = obj.learnParams{'verbose', 'values'};

Run = obj.configRun(Config);

assertEqual( Run.pc{1}, Run.pc{2} );
assertEqual( Run.pMax{1}, Run.pMax{2} );
assertEqual( Run.pMaxStat{1}, Run.pMaxStat{2} );

end

function testDefaultConfig(obj)
    
Run = obj.configRun(obj.config_default);
Run = [ obj.config_default Run ];
save('mmpc_default_run', 'Run');
obj.runAnalyse('default');

end

function testCustomConfig(obj)
    
Config = obj.config_default;
Config.netName = {'alarm'};
Config.nsmax = 10;
Config.targetVar = 37;
Config.sampleSize = 5000;
Config.rl = 1;
Run = obj.configRun(Config);
Run = [ Config Run ];
save('mmpc_custom_run', 'Run');
obj.runAnalyse('custom');

end

function testVarSs(obj)
    
nConfigs = size(obj.net, 1)*numel(obj.sampleParams{'sampleNo', 'values'})*numel(obj.sampleParams{'sampleSize', 'values'});

% replicate default config for each net, sample no, sample size and reuse level
Config = repmat(obj.config_default, nConfigs, 1);

n = 0;

for i=1:size(obj.net, 1) % for each net 
    for j=1:numel(obj.sampleParams{'sampleNo', 'values'}) % for each sample no
        for k=1:numel(obj.sampleParams{'sampleSize', 'values'}) % for each sample size
                
            n = n + 1;
                
            Config.netName{n} = obj.netParams{'netName', 'values'}{i};
            Config.nsmax(n) = obj.netParams{'nsmax', 'values'}(i);            
            Config.sampleNo(n) = obj.sampleParams{'sampleNo', 'values'}(j);                
            Config.sampleSize(n) = obj.sampleParams{'sampleSize', 'values'}(k);
            Config.includeNpc(n) = true; % include non-edge data
            
        end
    end
end

% debug
%Config = Config(1:2, :);
%nConfigs = 2;

Run = obj.configRun(Config);
Run = [ Config Run ];

save('mmpc_var_ss', 'Run');

obj.analyseVarSs();

end

end

methods(Static)

function Run = configRun(Config)
   
Config    
    
nConfigs = size(Config, 1);

% create runs dataset
time = zeros(nConfigs, 1);
pc = cell(nConfigs, 1);
pMax = cell(nConfigs, 1);
pMaxStat = cell(nConfigs, 1);

Run = dataset(time, pc, pMax, pMaxStat);

for i=1:nConfigs % for each configuration
    
    fprintf('\nRunning configuration %d of %d (%.2f%%)\n', i, nConfigs, i/nConfigs*100);

    a = [];
    
    % load sample
    load(sprintf('%s_sample_%d', Config.netName{i}, Config.sampleNo(i)), 'a');
    
    % precompute subscripts
    subs = double(a);

    % precompute number of levels
    nLevels = org.mensxmachina.commons.stats.nlevels(a);

    disp('Learning pc with mmpc...');

    % dataset of this sample size
    data_ss = a(1:Config.sampleSize(i), :);

    % subscripts of this sample size
    subs_ss = subs(1:Config.sampleSize(i), :);

    mmpc_params = { ...
        'alpha', Config.alpha(i), ...
        'nsmax', Config.nsmax(i), ...
        'rl', Config.rl(i), ...
        'includeNpc', Config.includeNpc(i), ...
        'verbose', Config.verbose(i), ...
        'subs', subs_ss, ...
        'nLevels', nLevels ...
        };
    
    tic;
    
    [pc pMax pMaxStat] = ...
        org.mensxmachina.bnet.mmpc( ...
            data_ss, Config.targetVar(i), mmpc_params{:} );

    time = toc;
    time
    
    pcInd = find(pc);

    % assert all edges have max p-values <= alpha
    assertTrue( all(pMax(pcInd) <= Config.alpha(i)) );

    % assert all non-edges have
    % either max p-values > alpha or NaN max p-values
    isNanNonPcPMax = isnan(pMax(~pcInd));
    assertTrue( all(pMax(~pcInd) > Config.alpha(i) | isNanNonPcPMax) );

    if any(isNanNonPcPMax)
        disp('Some NaN max p-values were obtained.');
    end
    
    % fill run data
    Run.time(i) = time;
    Run.pc{i} = pc;
    Run.pMax{i} = pMax;
    Run.pMaxStat{i} = pMaxStat;

end

end

function runAnalyse(runName)
Run = [];   
    
load( sprintf('mmpc_%s_run', runName), 'Run');

netName = Run.netName{1};

load( sprintf('mxm_bnet_%s_net', lower(netName)) , 'net');

% create performance columns

% get performance object
slp = org.mensxmachina.bnet.pcperf(net.skeleton, Run.pc{1}, Run.targetVar(1));

Run.sensitivity = slp.sensitivity;
Run.specificity = slp.specificity;

fprintf('\nSensitivity = %.2f%%\n', Run.sensitivity*100);
fprintf('\nSpecificity = %.2f%%\n', Run.specificity*100);

% save it with the new columns
save( sprintf('mmpc_%s_run', runName), 'Run');

org.mensxmachina.bnet.plotpcconfusion(net.skeleton, Run.pc{1}, Run.targetVar(1));

end

function analyseVarSs()  
%ANALYSEVARSS analyse variable sample size runs

Run = [];
    
load('mmpc_var_ss', 'Run');

netName = unique(Run.netName);

for i=1:numel(netName) % for each net
    % load net
    load( sprintf('mxm_bnet_%s_net', netName{i}), 'net');
    skeleton.(netName{i}) = net.skeleton;
end

nRuns = size(Run, 1);

% create performance columns

Run.sensitivity = zeros(nRuns, 1);
Run.specificity = zeros(nRuns, 1);
Run.auc = zeros(nRuns, 1);

for i=1:nRuns % for each run

    skeleton_true = skeleton.(Run.netName{i});
    
    % get performance object
    slp = org.mensxmachina.bnet.pcperf( ...
        skeleton_true, Run.pc{i}, Run.targetVar(i));
    
    Run.sensitivity(i) = slp.sensitivity;
    Run.specificity(i) = slp.specificity;
    
    Run.auc(i) = org.mensxmachina.bnet.pcauc( ...
        skeleton.(Run.netName{i}), Run.pMax{i}, Run.targetVar(i) );
    
end

% save it with the new columns
save('mmpc_var_ss', 'Run');

close all;

sampleSize = unique(Run.sampleSize);
sampleNo = unique(Run.sampleNo);

for i=1:numel(netName) % for each net
        
    pMax_ss = cell(1, numel(sampleSize));
    
    for j=1:numel(sampleSize) % for each sample size
        
        iNetjSs1stRun = Run( ...
            strcmp(Run.netName, netName{i}) & ...
            Run.sampleSize == sampleSize(j) ...
            & Run.sampleNo == sampleNo(1), : );
        
        pMax_ss{j} = iNetjSs1stRun.pMax{1};
        
    end
    
    figure;
    
    [tpr fpr threshold] = org.mensxmachina.bnet.plotpcroc( ...
        repmat({skeleton.(netName{i})}, 1, numel(sampleSize)), pMax_ss, repmat(iNetjSs1stRun.targetVar(1), numel(sampleSize)) );
    
    legend( arrayfun(@(i) num2str(i), sampleSize', 'UniformOutput', false) );
    set(gca, 'XTick', [0 iNetjSs1stRun.alpha(1) 1]);
    
    figureFilename = ...
            sprintf('mmpc_%s_roc_foreach_ss', netName{i});
    
    % print plot
    print(sprintf('-f%d', gcf), '-r300', '-depsc', figureFilename);    
    
end

PerfVar = TestMmpc.PerfVar;
PerfVar = PerfVar([1 4], :); % time and AUC only

for i=1:numel(netName) % for each net
    
    % select runs on this net
    Run_net = Run(strcmp(Run.netName, netName{i}), :);
    
    for j=1:size(PerfVar, 1) % for each performance variable
        
        figure;
        
        [h b] = org.mensxmachina.commons.stats.grpstatplot( ...
            Run_net, 'sampleSize', PerfVar.name{j}, 'errorstat', @std);

        % annotate plot
        theylim = ylim;
        ylim([ 0 theylim(2) ]);

        xlabel('Sample size');
        ylabel(sprintf('Mean %s', PerfVar.name{j}));
        title(sprintf('Mean %s learning %s vs sample size', PerfVar.label{j}, netName{i}));

        figureFilename = ...
            sprintf('mmpc_%s_mean_%s_vs_ss', ...
                netName{i}, PerfVar.name{j});

        % print plot
        print(sprintf('-f%d', gcf), '-r300', '-depsc', figureFilename);
        
    end
        
end
  
end

function PerfVar = PerfVar()
    PerfVar = dataset();
    PerfVar.name = {'time'; 'sensitivity'; 'specificity'; 'auc'};
    PerfVar.bestname = {'minimum execution time'; 'maximum sensitivity'; 'maximum specificity'; 'maximum AUC'};
    PerfVar.rlFun = {@TestMmpc.minrl; @TestMmpc.maxrl; @TestMmpc.maxrl; @TestMmpc.maxrl};
    PerfVar.label = {'time (s)'; 'sensitivity'; 'specificity'; 'AUC'};
    PerfVar.bestrl_ss = cell(4, 1);
end

function minrl = minrl(a)        
    [~, minrl] = min(a);
    minrl = minrl - 1;
end

function maxrl = maxrl(a)        
    [~, maxrl] = max(a);
    maxrl = maxrl - 1;
end
    
end

end